# coding: utf-8

from __future__ import print_function
import sys
import os
import re
from collections import OrderedDict
import lz4.frame
import gzip
import shutil


END_LINE_CH = "\n".encode()
INDEX_FILE_NAME = "BACKUP_INDEX.LST".encode()


def _log(msg, level="info"):
    print("[{level}]{msg}".format(level=level.upper(), msg=msg))


def restore_obs_file(src_file, target_dir):
    """
    Restore src_file to target_dir.
    Args:
        src_file: Full path to backup file download from obs.
        target_dir: Restored target directory. Must be an empty directory or
            does not exist.
    Returns:
        Return True if restore successfully, otherwise return False.
    """
    src_file = os.path.abspath(src_file)
    target_dir = os.path.abspath(target_dir)
    if not src_file or not os.path.exists(src_file):
        _log("The src_file(%s) is not exists." % src_file, "error")
        return False
    parent_dir = os.path.dirname(target_dir)
    if not os.path.isdir(parent_dir):
        _log("The parent of target_dir(%s) is not exists." % target_dir, "error")
        return False
    if os.path.isdir(target_dir) and os.listdir(target_dir):
        _log("Target_dir(%s) is not empty." % target_dir, "error")
        return False
    if not os.path.isdir(target_dir):
        os.mkdir(target_dir)
    temp_dir = os.path.join(target_dir, "temp")
    try:
        os.mkdir(temp_dir)
        index_file = os.path.join(temp_dir, "BACKUP_INDEX.LST")
        with open(src_file, "rb") as r_fd, open(index_file, "wb") as w_fd:
            offset = 0
            index_bytes = bytes()
            ch = r_fd.read(1)
            while ch and ch != END_LINE_CH:
                index_bytes += ch
                offset += 1
                w_fd.write(ch)
                ch = r_fd.read(1)
            if ch == END_LINE_CH:
                w_fd.write(ch)
                offset += 1
            pattern = r"^(\S+) (\d+)$".encode()
            pat_match = re.match(pattern, index_bytes)
            if not pat_match:
                _log("Parse index file from source file failed.", "error")
                return False
            if pat_match.group(1) != INDEX_FILE_NAME:
                _log("Parse index file from source file failed.", "error")
                return False
            file_size = int(pat_match.group(2))

            file_to_size_map = OrderedDict()
            got_bytes = offset
            line_bytes = bytes()
            while got_bytes < file_size:
                ch = r_fd.read(1)
                if not ch:
                    break
                got_bytes += 1
                line_bytes += ch
                w_fd.write(ch)

                if ch == END_LINE_CH:
                    pat_match = re.match(pattern, line_bytes)
                    if not pat_match:
                        _log("Parse index file from source file failed.", "error")
                        return False
                    file_to_size_map[pat_match.group(1).decode()] = \
                        int(pat_match.group(2))
                    line_bytes = bytes()
            tmp_buffer_size = 16 * 1024
            for name, size in file_to_size_map.items():
                output_file = os.path.join(temp_dir, name)
                with open(output_file, "wb") as out_fd:
                    got_bytes = 0
                    while size > got_bytes:
                        read_len = min(tmp_buffer_size, size-got_bytes)
                        read_bytes = r_fd.read(read_len)
                        if read_len != len(read_bytes):
                            _log("Parse data file from source file failed.", "error")
                            return False
                        out_fd.write(read_bytes)
                        got_bytes += read_len
        if not _decompress_files(target_dir):
            return False
    finally:
        if os.path.exists(temp_dir):
            _delete_dir(temp_dir)
    _log("Parse success.", "info")
    return True


def _decompress_files(target_dir):
    temp_dir = os.path.join(target_dir, "temp")
    for filename in sorted(os.listdir(temp_dir), reverse=True):
        abs_filename = os.path.join(temp_dir, filename)
        if filename == "base.tar.gz":
            if not _decompress_data(target_dir, abs_filename):
                return False
        elif re.match(r"^\d{1,10}\.tar.gz$", filename):
            if not _decompress_tblspc(target_dir, abs_filename):
                return False
        elif re.match(r"^[A-F0-9]{24}\.gz$", filename):
            if not _decompress_xlog(target_dir, abs_filename):
                return False
        elif re.match(r"^[A-F0-9]{8}\.history$", filename):
            if not _restore_history_file(target_dir, abs_filename):
                return False
    return True


def __delete_file(filename):
    cmd = 'rm -f %s' % filename
    if os.system(cmd) != 0:
        _log("Execute command(%s) failed." % cmd, "error")
        return False
    return True


def _delete_dir(dirname):
    cmd = 'rm -rf %s' % dirname
    if os.system(cmd) != 0:
        _log("Execute command(%s) failed." % cmd, "error")
        return False
    return True

def _copy_file(src, tar):
    cmd = 'cp -f %s %s' % (src, tar)
    if os.system(cmd) != 0:
        _log("Execute command(%s) failed." % cmd, "error")
        return False
    return True

def _decompress_data(target_dir, filename):
    data_dir = os.path.join(target_dir, "data")
    if not os.path.isdir(data_dir):
        os.mkdir(data_dir)
    extract_command = "tar zxf %s -C %s" % (filename, data_dir)
    if os.system(extract_command) != 0:
        _log("Execute command(%s) failed." % extract_command, "error")
        return False

    # recovery.done文件中包含密碼等敏感信息，需要刪除
    recovery_path = os.path.join(data_dir, "recovery.done")
    if not __delete_file(recovery_path):
        return False
    return True

def _decompress_xlog(target_dir, filename):
    xlog_dir = os.path.join(target_dir, "xlog")
    if not os.path.isdir(xlog_dir):
        os.mkdir(xlog_dir)
    try:
        with lz4.frame.open(filename, "rb") as r_fd, \
                open(filename[:-3], "wb") as w_fd:
            w_fd.write(r_fd.read())
            w_fd.flush()
            os.fsync(w_fd)
    except Exception:
        try:
            with gzip.open(filename, "rb") as r_fd, \
                    open(filename[:-3], "wb") as w_fd:
                shutil.copyfileobj(r_fd, w_fd)
        except Exception as err_msg:
            _log(
                "Decompress xlog(%s) failed. Error Msg: %s" % (
                    filename, err_msg),
                "error"
            )
            return False
    if not _copy_file(filename[:-3], xlog_dir):
        return False
    return True

def _restore_history_file(target_dir, src_file):
    xlog_dir = os.path.join(target_dir, "xlog")
    if not os.path.isdir(xlog_dir):
        os.mkdir(xlog_dir)
    file_name = os.path.basename(src_file)
    target_f = os.path.join(xlog_dir, file_name)
    if not _copy_file(src_file, target_f):
        return False
    return True

def _decompress_tblspc(target_dir, filename):
    tblspc_dir = os.path.join(target_dir, "tblspc")
    tblspc_oid = os.path.basename(filename).partition(".")[0]
    tblspc_path = os.path.join(tblspc_dir, tblspc_oid)
    if not os.path.isdir(tblspc_path):
        os.makedirs(tblspc_path)
    extract_command = "tar zxf %s -C %s" % (filename, tblspc_path)
    if os.system(extract_command) != 0:
        _log("Execute command(%s) failed." % extract_command, "error")
        return False
    return True


def print_help():
    help_msg = """usage: python src_file target_dir.
    src_file: Full path to backup file download from obs.
    target_dir: Restored target directory. Must be an empty directory or does not exist.
    """
    print(help_msg)


if __name__ == "__main__":
    if len(sys.argv) < 3:
        print_help()
        exit(0)
    restore_obs_file(sys.argv[1], sys.argv[2])
